"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
/**
 * The cache entry is a typed container of cache data used to track the
 * creation and expiration of cache entries.
 */
class CacheEntry {
  /**
   * Initializes the cache entry.
   *
   * @param {*} value The cache entry value.
   * @param {number} ttl The time to live in milliseconds.
   */
  constructor(value, ttl) {
    /**
     * Cache entry value.
     *
     * @type {*}
     */
    this._value = value;

    /**
     * The time to live in milliseconds. The cache entry is considered
     * expired after this time.
     *
     * @type {number}
     */
    this._ttl = ttl;

    /**
     * The timestamp of creation of this cache entry.
     *
     * @type {number}
     */
    this._created = Date.now();
  }

  /**
   * Returns `true` if this entry has expired.
   *
   * @return {boolean} `true` if this entry has expired.
   */
  isExpired() {
    let now = Date.now();
    return now > this._created + this._ttl;
  }

  /**
   * Exports this cache entry into a JSON-serializable object.
   *
   * @return {{value: *, ttl: number}} This entry exported to a
   *         JSON-serializable object.
   */
  serialize() {
    return { value: this._value, ttl: this._ttl };
  }

  /**
   * Returns the entry value.
   *
   * @return {*} The entry value.
   */
  getValue() {
    return this._value;
  }
}
exports.default = CacheEntry;

typeof $IMA !== 'undefined' && $IMA !== null && $IMA.Loader && $IMA.Loader.register('ima/cache/CacheEntry', [], function (_export, _context) {
 'use strict';
 return {
   setters: [],
   execute: function () {
     _export('default', exports.default);
   }
 };
});
