'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _MapStorage = require('./MapStorage');

var _MapStorage2 = _interopRequireDefault(_MapStorage);

var _SessionStorage = require('./SessionStorage');

var _SessionStorage2 = _interopRequireDefault(_SessionStorage);

var _Storage = require('./Storage');

var _Storage2 = _interopRequireDefault(_Storage);

var _CacheEntry = require('../cache/CacheEntry');

var _CacheEntry2 = _interopRequireDefault(_CacheEntry);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * The {@codelink SessionMap} storage is an implementation of the
 * {@codelink Storage} interface acting as a synchronization proxy between
 * the underlying map storage and the {@code sessionStorage} DOM storage.
 */
class SessionMapStorage extends _Storage2.default {
  static get $dependencies() {
    return [_MapStorage2.default, _SessionStorage2.default];
  }

  /**
   * Initializes the storage.
   *
   * @param {MapStorage} map The map storage to use.
   * @param {SessionStorage} session The session storage to use.
   */
  constructor(map, session) {
    super();

    /**
     * The map storage, synced with the session storage.
     *
     * @type {MapStorage}
     */
    this._map = map;

    /**
     * The session storage, synced with the map storage.
     *
     * @type {SessionStorage}
     */
    this._session = session;
  }

  /**
   * @inheritdoc
   */
  init() {
    this._map.clear();
    for (let key of this._session.keys()) {
      this._map.set(key, this._session[key]);
    }

    return this;
  }

  /**
   * @inheritdoc
   */
  has(key) {
    return this._map.has(key);
  }

  /**
   * @inheritdoc
   */
  get(key) {
    return this._map.get(key);
  }

  /**
   * @inheritdoc
   */
  set(key, value) {
    let canBeSerializedToJSON = !(value instanceof Promise) && (!(value instanceof _CacheEntry2.default) || !(value.getValue() instanceof Promise));
    if (canBeSerializedToJSON) {
      this._session.set(key, value);
    }

    this._map.set(key, value);
    return this;
  }

  /**
   * @inheritdoc
   */
  delete(key) {
    this._session.delete(key);
    this._map.delete(key);
    return this;
  }

  /**
   * @inheritdoc
   */
  clear() {
    this._session.clear();
    this._map.clear();
    return this;
  }

  /**
   * @inheritdoc
   */
  keys() {
    return this._map.keys();
  }

  /**
   * @override
   */
  size() {
    return this._map.size();
  }
}
exports.default = SessionMapStorage;

typeof $IMA !== 'undefined' && $IMA !== null && $IMA.Loader && $IMA.Loader.register('ima/storage/SessionMapStorage', [], function (_export, _context) {
 'use strict';
 return {
   setters: [],
   execute: function () {
     _export('default', exports.default);
   }
 };
});
